from __future__ import annotations

from math import floor
from typing import TYPE_CHECKING, Callable, Hashable, Literal

import numpy
import pandas

import convoys.multi

if TYPE_CHECKING:
    pass

__all__ = ["export_cohorts"]


_models: dict[str, Callable[[bool], convoys.multi.MultiModel]] = {
    "kaplan-meier": lambda _: convoys.multi.KaplanMeier(),
    "exponential": lambda ci: convoys.multi.Exponential(mcmc=ci),
    "weibull": lambda ci: convoys.multi.Weibull(mcmc=ci),
    "gamma": lambda ci: convoys.multi.Gamma(mcmc=ci),
    "generalized-gamma": lambda ci: convoys.multi.GeneralizedGamma(mcmc=ci),
}


def export_cohorts(
    G: numpy.ndarray,
    B: numpy.ndarray,
    T: numpy.ndarray,
    t_max: int | float | None = None,
    model: Literal[
        "kaplan-meier", "exponential", "weibull", "gamma", "generalized-gamma"
    ]
    | convoys.multi.MultiModel = "kaplan-meier",
    ci: float | None = None,
    groups: list[Hashable] | None = None,
    specific_groups: list[Hashable] | None = None,
) -> pandas.DataFrame:
    """Helper function to fit data using a model and then
    export the model predictions as a DataFrame. The Dataframe will
    have the columns "group", "t", and "prediction_value"; if ci is not None,
    it will also have the columns "ci_low" and "ci_high". "t" will be integers
    in the range from 0 to t_max (or max(T) if t_max is None). This
    format enables easy storage in a database and plotting with BI tools.

    :param G: numpy array of shape :math:`n`, containing integers representing group
        assignments.
    :param B: numpy array of shape :math:`n`, containing booleans representing
        whether or not the subject 'converted' at time delta :math:`t`.
    :param T: numpy array of shape :math:`n`, containing floats representing the
        time delta :math:`t` between creation and either conversion or censoring.
    :param t_max: (optional) max value for x axis
    :param model: (optional, default is kaplan-meier) model to fit.
        Can be an instance of :class:`multi.MultiModel` or a string
        identifying the model. One of 'kaplan-meier', 'exponential',
        'weibull', 'gamma', or 'generalized-gamma'.
    :param ci: confidence interval, value from 0-1, or None (default) if
        no confidence interval is to be plotted
    :param groups: list of group labels
    :param specific_groups: subset of groups to plot

    See  :meth:`convoys.utils.get_arrays` which is handy for converting
    a Pandas dataframe into arrays `G`, `B`, `T`.
    """

    if model not in _models.keys():
        if not isinstance(model, convoys.multi.MultiModel):
            raise ValueError("model incorrectly specified")

    if groups is None:
        groups = list(set(G))

    # Set x scale
    if t_max is None:
        t_max = max(T)
    if not isinstance(model, convoys.multi.MultiModel):
        # Fit model
        m = _models[model](bool(ci))
        m.fit(G, B, T)
    else:
        m = model

    if specific_groups is None:
        specific_groups = groups

    if len(set(specific_groups).intersection(groups)) != len(specific_groups):
        raise ValueError("specific_groups not a subset of groups!")

    RESULT_LENGTH = floor(t_max + 1)
    t = numpy.arange(RESULT_LENGTH)
    data: list[pandas.DataFrame] = []

    for group in specific_groups:
        j = groups.index(group)  # matching index of group

        if ci is not None:
            result = m.predict_ci(j, t, ci=ci)
            result_df = pandas.DataFrame(
                data=result, columns=["prediction_value", "ci_low", "ci_high"]
            )
        else:
            result = m.predict(j, t)
            result_df = pandas.DataFrame(data=result, columns=["prediction_value"])
        result_df["t"] = t
        result_df["group"] = group
        data.append(result_df)

    unioned_data = pandas.concat(data)
    return unioned_data



# -*- coding: utf-8 -*-
#
# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

sys.path.insert(0, os.path.abspath(os.path.pardir))


# -- Project information -----------------------------------------------------

project = "Convoys"
copyright = "2018, Erik Bernhardsson"  # noqa: A001
author = "Erik Bernhardsson"

# The short X.Y version
version = ""
# The full version, including alpha/beta/rc tags
release = ""


# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.doctest",
    "sphinx.ext.todo",
    "sphinx.ext.coverage",
    "sphinx.ext.mathjax",
    "sphinx.ext.viewcode",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path .
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "alabaster"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {
    "github_user": "tconbeer",
    "github_repo": "convoys2",
    "github_button": True,
    "github_type": "star",
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.
#
html_sidebars = {
    "**": [
        "about.html",
        "localtoc.html",
        "searchbox.html",
    ]
}

# Make it possible to build docs without dependencies
autodoc_mock_imports = [
    "autograd",
    "autograd_gamma",
    "emcee",
    "matplotlib",
    "numpy",
    "pandas",
    "progressbar",
    "scipy",
]

# -- Extension configuration -------------------------------------------------

# -- Options for todo extension ----------------------------------------------

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True



from types import TracebackType
from typing import Callable, Protocol, TypeVar

from .flaky_pytest_plugin import FlakyPlugin

T_args_contra = TypeVar("T_args_contra", contravariant=True)
T_kwargs_contra = TypeVar("T_kwargs_contra", contravariant=True)
T_ret_co = TypeVar("T_ret_co", covariant=True)

class WrappedFunc(Protocol[T_args_contra, T_kwargs_contra, T_ret_co]):
    def __call__(self, *args: T_args_contra, **kwds: T_kwargs_contra) -> T_ret_co: ...

def flaky(
    max_runs: WrappedFunc | int | None = None,
    min_passes: int | None = None,
    rerun_filter: Callable[
        [
            tuple[type[BaseException], BaseException, TracebackType]
            | tuple[None, None, None],
            str,
            Callable,
            FlakyPlugin,
        ],
        bool,
    ]
    | None = None,
) -> WrappedFunc | Callable[[WrappedFunc], T_ret_co]: ...



import datetime
import random

import matplotlib
import numpy
import pandas
import pytest
import scipy.stats  # type: ignore[import-untyped]

matplotlib.use("Agg")  # Needed for matplotlib to run in Travis


class Utilities:
    @staticmethod
    def sample_weibull(k: float, lambd: float) -> float:
        # scipy.stats is garbage for this
        # exp(-(x * lambda)^k) = y
        base: numpy.float64 = -numpy.log(random.random())
        return base ** (1.0 / k) / lambd

    @staticmethod
    def generate_censored_data(
        N: numpy.ndarray, E: numpy.ndarray, C: numpy.ndarray
    ) -> tuple[numpy.ndarray, numpy.ndarray]:
        B = numpy.array([c and e < n for n, e, c in zip(N, E, C, strict=False)])
        T = numpy.array([e if b else n for e, b, n in zip(E, B, N, strict=False)])
        return B, T


@pytest.fixture(scope="session")
def utilities() -> Utilities:
    return Utilities()


def generate_weibull_df() -> pandas.DataFrame:
    cs = (0.3, 0.5, 0.7)
    k = 0.5
    lambd = 0.1
    n = 1000
    groups = [r % len(cs) for r in range(n)]
    C = numpy.array([bool(random.random() < cs[g]) for g in groups])
    N = scipy.stats.expon.rvs(scale=10.0 / lambd, size=(n,))
    E = numpy.array([Utilities.sample_weibull(k, lambd) for r in range(n)])
    B, T = Utilities.generate_censored_data(N, E, C)

    def x2t(x: int) -> datetime.datetime:
        return datetime.datetime(2000, 1, 1) + datetime.timedelta(days=x)

    return pandas.DataFrame(
        data=dict(
            group=["Group %d" % g for g in groups],
            created=[x2t(0) for g in groups],
            converted=[x2t(t) if b else None for t, b in zip(T, B, strict=False)],
            now=[x2t(n) for n in N],
        )
    )


weibull_data = generate_weibull_df()


@pytest.fixture(scope="function")
def weibull_df() -> pandas.DataFrame:
    return weibull_data.copy()



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Hashable, Literal

import numpy
from matplotlib import pyplot

import convoys.multi

if TYPE_CHECKING:
    from matplotlib.axes import Axes

__all__ = ["plot_cohorts"]


_models: dict[str, Callable[[bool], convoys.multi.MultiModel]] = {
    "kaplan-meier": lambda _: convoys.multi.KaplanMeier(),
    "exponential": lambda ci: convoys.multi.Exponential(mcmc=ci),
    "weibull": lambda ci: convoys.multi.Weibull(mcmc=ci),
    "gamma": lambda ci: convoys.multi.Gamma(mcmc=ci),
    "generalized-gamma": lambda ci: convoys.multi.GeneralizedGamma(mcmc=ci),
}


def plot_cohorts(
    G: numpy.ndarray,
    B: numpy.ndarray,
    T: numpy.ndarray,
    t_max: int | float | None = None,
    model: Literal[
        "kaplan-meier", "exponential", "weibull", "gamma", "generalized-gamma"
    ]
    | convoys.multi.MultiModel = "kaplan-meier",
    ci: float | None = None,
    ax: "Axes" | None = None,
    plot_kwargs: dict[str, Any] | None = None,
    plot_ci_kwargs: dict[str, Any] | None = None,
    groups: list[Hashable] | None = None,
    specific_groups: list[Hashable] | None = None,
    label_fmt: str = "%(group)s (n=%(n).0f, k=%(k).0f)",
) -> convoys.multi.MultiModel:
    """Helper function to fit data using a model and then plot the cohorts.

    :param G: numpy array of shape :math:`n`, containing integers representing group
        assignments.
    :param B: numpy array of shape :math:`n`, containing booleans representing
        whether or not the subject 'converted' at time t.
    :param T: numpy array of shape :math:`n`, containing floats representing the
        time delta between creation and either conversion or censoring.
    :param t_max: (optional) max value for x axis
    :param model: (optional, default is kaplan-meier) model to fit.
        Can be an instance of :class:`multi.MultiModel` or a string
        identifying the model. One of 'kaplan-meier', 'exponential',
        'weibull', 'gamma', or 'generalized-gamma'.
    :param ci: confidence interval, value from 0-1, or None (default) if
        no confidence interval is to be plotted
    :param ax: custom pyplot axis to plot on
    :param plot_kwargs: extra arguments to pyplot for the lines
    :param plot_ci_kwargs: extra arguments to pyplot for the confidence
        intervals
    :param groups: list of group labels
    :param specific_groups: subset of groups to plot
    :param label_fmt: custom format for the labels to use in the legend

    See  :meth:`convoys.utils.get_arrays` which is handy for converting
    a Pandas dataframe into arrays `G`, `B`, `T`.
    """

    if model not in _models.keys():
        if not isinstance(model, convoys.multi.MultiModel):
            raise ValueError("model incorrectly specified")

    if groups is None:
        groups = list(set(G))

    if ax is None:
        ax = pyplot.gca()

    # Set x scale
    if t_max is None:
        _, t_max = ax.get_xlim()
        t_max = max(t_max, max(T))
    if not isinstance(model, convoys.multi.MultiModel):
        # Fit model
        m = _models[model](bool(ci))
        m.fit(G, B, T)
    else:
        m = model

    if specific_groups is None:
        specific_groups = groups

    if len(set(specific_groups).intersection(groups)) != len(specific_groups):
        raise ValueError("specific_groups not a subset of groups!")

    # Plot
    t = numpy.linspace(0, t_max, 1000)  # type: ignore[arg-type]
    _, y_max = ax.get_ylim()
    # Reset to first color
    ax.set_prop_cycle(None)  # type:ignore[call-overload]
    for group in specific_groups:
        j = groups.index(group)  # matching index of group

        n = numpy.sum(G == j)
        k = numpy.sum(B[G == j])
        label = label_fmt % dict(group=group, n=n, k=k)

        if ci is not None:
            p_y, p_y_lo, p_y_hi = m.predict_ci(j, t, ci=ci).T
            merged_plot_ci_kwargs = {"alpha": 0.2}
            if plot_ci_kwargs is not None:
                merged_plot_ci_kwargs.update(plot_ci_kwargs)
            p = ax.fill_between(
                t,
                100.0 * p_y_lo,
                100.0 * p_y_hi,
                **merged_plot_ci_kwargs,  # type: ignore[arg-type]
            )
            color = p.get_facecolor()[0]  # reuse color for the line
        else:
            p_y = m.predict(j, t).T
            color = None

        merged_plot_kwargs = {"color": color, "linewidth": 1.5, "alpha": 0.7}
        if plot_kwargs is not None:
            merged_plot_kwargs.update(plot_kwargs)
        ax.plot(t, 100.0 * p_y, label=label, **merged_plot_kwargs)  # type: ignore[arg-type]
        y_max = max(y_max, 110.0 * max(p_y))

    ax.set_xlim(0, t_max)
    ax.set_ylim(0, y_max)
    ax.set_ylabel("Conversion rate %")
    ax.grid(True)
    return m



from pathlib import Path

import pandas
from matplotlib import pyplot

import convoys.plotting
import convoys.utils


def run() -> None:
    print("loading data")
    df = pandas.read_pickle("examples/marriage.pickle")
    df = df.sample(1000)  # speed up
    print(df)

    _, groups, (G, B, T) = convoys.utils.get_arrays(
        df, groups="sex", created="born", converted="married"
    )

    pyplot.figure(figsize=(6, 6))
    convoys.plotting.plot_cohorts(G, B, T, model="generalized-gamma", groups=groups)
    pyplot.legend()
    pyplot.xlabel("Age of marriage")
    convoys.plotting.plot_cohorts(
        G, B, T, model="kaplan-meier", groups=groups, plot_kwargs={"linestyle": "--"}
    )
    fig_path = Path(__file__).parent / "marriage-combined.png"
    pyplot.savefig(fig_path)


if __name__ == "__main__":
    run()



import datetime
import random
from pathlib import Path
from typing import TYPE_CHECKING, Literal

import flaky
import matplotlib
import numpy
import pandas
import pytest
import scipy.stats  # type: ignore[import-untyped]
from numpy.typing import ArrayLike

matplotlib.use("Agg")  # Needed for matplotlib to run in Travis
import convoys
import convoys.multi
import convoys.plotting
import convoys.regression
import convoys.single
import convoys.utils

if TYPE_CHECKING:
    from tests.conftest import Utilities


def test_kaplan_meier_model() -> None:
    data = [(2, 0), (3, 0), (6, 1), (6, 1), (7, 1), (10, 0)]
    now = pandas.Timestamp("2019-01-22")  # fix now end date for easier testing
    created_array = [now - pandas.DateOffset(n=t) for t, _ in data]
    converted_array = [
        ts + pandas.DateOffset(n=t) if e == 1 else numpy.nan
        for ts, (t, e) in zip(created_array, data, strict=False)
    ]
    df = pandas.DataFrame(
        {"created_at": created_array, "converted_at": converted_array, "group": 1}
    )
    df["now"] = now
    unit, groups, (G, B, T) = convoys.utils.get_arrays(
        df, converted="converted_at", created="created_at", unit="days"
    )
    m = convoys.multi.KaplanMeier()
    m.fit(G, B, T)
    assert m.predict(0, 9) == 0.75


def test_output_shapes(
    utilities: "Utilities",
    c: float = 0.3,
    lambd: float = 0.1,
    n: int = 1000,
    k: int = 5,
) -> None:
    X = numpy.random.randn(n, k)
    C = scipy.stats.bernoulli.rvs(c, size=(n,))
    N = scipy.stats.uniform.rvs(scale=5.0 / lambd, size=(n,))
    E = scipy.stats.expon.rvs(scale=1.0 / lambd, size=(n,))
    B, T = utilities.generate_censored_data(N, E, C)

    # Fit model with ci
    model = convoys.regression.Exponential(mcmc=True)
    model.fit(X, B, T)

    # Generate output without ci
    assert model.predict(X[0], 0).shape == ()
    assert model.predict([X[0], X[1]], 0).shape == (2,)
    assert model.predict([X[0]], [0, 1, 2, 3]).shape == (4,)
    assert model.predict([X[0], X[1], X[2]], [0, 1, 2]).shape == (3,)
    assert model.predict([[X[0], X[1]]], [[0], [1], [2]]).shape == (3, 2)
    assert model.predict([[X[0]], [X[1]]], [[0, 1, 2]]).shape == (2, 3)

    # Generate output with ci (same as above plus (3,))
    assert model.predict_ci(X[0], 0, ci=0.8).shape == (3,)
    assert model.predict_ci([X[0], X[1]], 0, ci=0.8).shape == (2, 3)
    assert model.predict_ci([X[0]], [0, 1, 2, 3], ci=0.8).shape == (4, 3)
    assert model.predict_ci([X[0], X[1], X[2]], [0, 1, 2], ci=0.8).shape == (3, 3)
    assert model.predict_ci([[X[0], X[1]]], [[0], [1], [2]], ci=0.8).shape == (3, 2, 3)
    assert model.predict_ci([[X[0]], [X[1]]], [[0, 1, 2]], ci=0.8).shape == (2, 3, 3)

    # Fit model without ci (should be the same)
    model = convoys.regression.Exponential(mcmc=False)
    model.fit(X, B, T)
    assert model.predict(X[0], 0).shape == ()
    assert model.predict([X[0], X[1]], [0, 1]).shape == (2,)


@flaky.flaky
def test_exponential_regression_model(
    utilities: "Utilities", c: float = 0.3, lambd: float = 0.1, n: int = 10000
) -> None:
    X = numpy.ones((n, 1))
    C: numpy.ndarray = scipy.stats.bernoulli.rvs(c, size=(n,))  # did it convert
    N: numpy.ndarray = scipy.stats.uniform.rvs(scale=5.0 / lambd, size=(n,))  # time now
    E: numpy.ndarray = scipy.stats.expon.rvs(
        scale=1.0 / lambd, size=(n,)
    )  # time of event
    B, T = utilities.generate_censored_data(N, E, C)
    model = convoys.regression.Exponential(mcmc=True)
    model.fit(X, B, T)
    assert 0.80 * c < model.predict([1], float("inf")) < 1.30 * c
    for t in [1, 3, 10]:
        d = 1 - numpy.exp(-lambd * t)
        assert 0.80 * c * d < model.predict([1], t) < 1.30 * c * d

    # Check the confidence intervals
    assert model.predict_ci([1], float("inf"), ci=0.95).shape == (3,)
    assert model.predict_ci([1], [0, 1, 2, 3], ci=0.95).shape == (4, 3)
    y, y_lo, y_hi = model.predict_ci([1], float("inf"), ci=0.95)
    assert 0.80 * c < y < 1.30 * c

    # Check the random variates
    will_convert, convert_at = model.rvs([1], n_curves=10000, n_samples=1)
    assert 0.80 * c < numpy.mean(will_convert) < 1.30 * c
    convert_times = convert_at[will_convert]
    for t in [1, 3, 10]:
        d = 1 - numpy.exp(-lambd * t)
        assert 0.70 * d < (convert_times < t).mean() < 1.30 * d

    # Fit a linear model
    model = convoys.regression.Exponential(mcmc=False, flavor="linear")
    model.fit(X, B, T)
    model_c = model.params["map"]["b"] + model.params["map"]["beta"][0]
    assert 0.9 * c < model_c < 1.1 * c
    for t in [1, 3, 10]:
        d = 1 - numpy.exp(-lambd * t)
        assert 0.80 * c * d < model.predict([1], t) < 1.30 * c * d


@flaky.flaky
def test_weibull_regression_model(
    utilities: "Utilities",
    cs: tuple[float, float, float] = (0.3, 0.5, 0.7),
    lambd: float = 0.1,
    k: float = 0.5,
    n: int = 10000,
) -> None:
    X = numpy.array([[r % len(cs) == j for j in range(len(cs))] for r in range(n)])
    C = numpy.array([bool(random.random() < cs[r % len(cs)]) for r in range(n)])
    N = scipy.stats.uniform.rvs(scale=5.0 / lambd, size=(n,))
    E = numpy.array([utilities.sample_weibull(k, lambd) for r in range(n)])
    B, T = utilities.generate_censored_data(N, E, C)

    model = convoys.regression.Weibull()
    model.fit(X, B, T)

    # Validate shape of results
    x: "ArrayLike" = numpy.ones((len(cs),))
    assert model.predict(x, float("inf")).shape == ()
    assert model.predict(x, 1).shape == ()
    assert model.predict(x, [1, 2, 3, 4]).shape == (4,)

    # Check results
    for r, c in enumerate(cs):
        x = [int(r == j) for j in range(len(cs))]
        assert 0.80 * c < model.predict(x, float("inf")) < 1.30 * c

    # Fit a linear model
    model = convoys.regression.Weibull(mcmc=False, flavor="linear")
    model.fit(X, B, T)
    model_cs = model.params["map"]["b"] + model.params["map"]["beta"]
    for model_c, c in zip(model_cs, cs, strict=False):
        assert 0.8 * c < model_c < 1.2 * c


@flaky.flaky
def test_gamma_regression_model(
    utilities: "Utilities",
    c: float = 0.3,
    lambd: float = 0.1,
    k: float = 3.0,
    n: int = 10000,
) -> None:
    # TODO: this one seems very sensitive to large values for N (i.e. less censoring)
    X = numpy.ones((n, 1))
    C = scipy.stats.bernoulli.rvs(c, size=(n,))
    N = scipy.stats.uniform.rvs(scale=20.0 / lambd, size=(n,))
    E = scipy.stats.gamma.rvs(a=k, scale=1.0 / lambd, size=(n,))
    B, T = utilities.generate_censored_data(N, E, C)

    model = convoys.regression.Gamma()
    model.fit(X, B, T)
    assert 0.80 * c < model.predict([1], float("inf")) < 1.30 * c
    assert 0.80 * k < numpy.mean(model.params["map"]["k"]) < 1.30 * k

    # Fit a linear model
    model = convoys.regression.Gamma(mcmc=False, flavor="linear")
    model.fit(X, B, T)
    model_c = model.params["map"]["b"] + model.params["map"]["beta"][0]
    assert 0.9 * c < model_c < 1.1 * c


@flaky.flaky
def test_linear_model(
    utilities: "Utilities",
    n: int = 10000,
    m: int = 5,
    k: float = 3.0,
    lambd: float = 0.1,
) -> None:
    # Generate data with little censoring
    # The coefficients should be quite close to their actual value
    cs = numpy.random.dirichlet(numpy.ones(m))
    X = numpy.random.binomial(n=1, p=0.5, size=(n, m))
    C = numpy.random.rand(n) < numpy.dot(X, cs.T)
    N = scipy.stats.uniform.rvs(scale=20.0 / lambd, size=(n,))
    E = numpy.array([utilities.sample_weibull(k, lambd) for r in range(n)])
    B, T = utilities.generate_censored_data(N, E, C)

    model = convoys.regression.Weibull(mcmc=False, flavor="linear")
    model.fit(X, B, T)

    # Check the fitted parameters
    model_cs = model.params["map"]["b"] + model.params["map"]["beta"]
    for model_c, c in zip(model_cs, cs, strict=False):
        assert c - 0.03 < model_c < c + 0.03
    model_lambds = numpy.exp(model.params["map"]["a"] + model.params["map"]["alpha"])
    for model_lambd in model_lambds:
        assert 0.95 * lambd < model_lambd < 1.05 * lambd

    # Check predictions
    for i, c in enumerate(cs):
        x = numpy.array([float(j == i) for j in range(m)])
        p = model.predict(x, float("inf"))
        assert c - 0.03 < p < c + 0.03
        t = 10.0
        p = model.predict(x, t)
        f = 1 - numpy.exp(-((t * lambd) ** k))
        assert c * f - 0.03 < p < c * f + 0.03


@flaky.flaky
def test_exponential_pooling(
    utilities: "Utilities",
    c: "ArrayLike" = 0.5,
    lambd: float = 0.01,
    n: int = 10000,
    ks: list[int] | tuple[int, ...] = (1, 2, 3),
) -> None:
    # Generate one series of n observations with c conversion rate
    # Then k1...kn series with zero conversion
    # The predicted conversion rates should go towards c for the small cohorts
    G = numpy.zeros(n + sum(ks))
    C = numpy.zeros(n + sum(ks))
    N = numpy.zeros(n + sum(ks))
    E = numpy.zeros(n + sum(ks))
    offset = 0
    for i, k in enumerate([n] + list(ks)):
        G[offset : offset + k] = i
        offset += k
    C[:n] = scipy.stats.bernoulli.rvs(c, size=(n,))
    N[:] = 1000.0
    E[:n] = scipy.stats.expon.rvs(scale=1.0 / lambd, size=(n,))
    B, T = utilities.generate_censored_data(N, E, C)

    # Fit model
    model = convoys.multi.Exponential()
    model.fit(G, B, T)

    # Generate predictions for each cohort
    c = numpy.array([model.predict(i, float("inf")) for i in range(1 + len(ks))])
    assert numpy.all(c[1:] > 0.25)  # rough check
    assert numpy.all(c[1:] < 0.50)  # same
    assert numpy.all(numpy.diff(c) < 0)  # c should be monotonically decreasing


def test_convert_dataframe(weibull_df: pandas.DataFrame) -> None:
    df = weibull_df
    unit, groups, (G, B, T) = convoys.utils.get_arrays(df)
    assert G.shape == B.shape == T.shape == (len(df),)


def test_convert_dataframe_features(weibull_df: pandas.DataFrame) -> None:
    df = weibull_df
    df["features"] = [
        tuple(numpy.random.randn() for z in range(3)) for g in df["group"]
    ]
    df = df.drop("group", axis=1)
    unit, groups, (X, B, T) = convoys.utils.get_arrays(df)
    assert X.shape == (len(df), 3)


def test_convert_dataframe_features_multi_cols(weibull_df: pandas.DataFrame) -> None:
    # Generate from multiple columns
    df = weibull_df
    df["feature_1"] = [numpy.random.randn() for g in df["group"]]
    df["feature_2"] = [numpy.random.randn() for g in df["group"]]
    df = df.drop("group", axis=1)
    unit, groups, (X, B, T) = convoys.utils.get_arrays(
        df, features=("feature_1", "feature_2")
    )
    assert X.shape == (len(df), 2)


def test_convert_dataframe_infer_now(weibull_df: pandas.DataFrame) -> None:
    df = weibull_df
    df = df.drop("now", axis=1)

    unit, groups, (G1, B1, T1) = convoys.utils.get_arrays(df, unit="days")

    # Now, let's make the timezone-naive objects timezone aware
    utc = datetime.timezone.utc
    local = datetime.datetime.now(utc).astimezone().tzinfo
    df[["created", "converted"]] = df[["created", "converted"]].map(
        lambda z: z.replace(tzinfo=local)
    )
    unit, groups, (G2, B2, T2) = convoys.utils.get_arrays(df, unit="days")

    # Convert everything to UTC and make sure it's still the same
    df[["created", "converted"]] = df[["created", "converted"]].map(
        lambda z: z.tz_convert(utc)
    )
    unit, groups, (G3, B3, T3) = convoys.utils.get_arrays(df, unit="days")

    # Let's check that all deltas are the same
    # There will be some slight clock drift, so let's accept up to 3s
    for t1, t2, t3 in zip(T1, T2, T3, strict=False):
        assert 0 <= t2 - t1 < 3.0 / (24 * 60 * 60)
        assert 0 <= t3 - t1 < 3.0 / (24 * 60 * 60)


def test_convert_dataframe_timedeltas(weibull_df: pandas.DataFrame) -> None:
    df = weibull_df

    unit, groups, (G1, B1, T1) = convoys.utils.get_arrays(df, unit="days")
    df2 = pandas.DataFrame(
        {
            "group": df["group"],
            "converted": df["converted"] - df["created"],
            "now": df["now"] - df["created"],
        }
    )
    unit, groups, (G2, B2, T2) = convoys.utils.get_arrays(df2, unit="days")

    for t1, t2 in zip(T1, T2, strict=False):
        assert 0 <= t2 - t1 < 3.0 / (24 * 60 * 60)


def test_convert_dataframe_more_args(weibull_df: pandas.DataFrame) -> None:
    df = weibull_df
    unit, groups, (G, B, T) = convoys.utils.get_arrays(df, max_groups=2)
    assert groups is not None
    assert len(groups) <= 2
    unit, groups, (G, B, T) = convoys.utils.get_arrays(df, group_min_size=9999)
    assert G.shape == (0,)


def test_convert_dataframe_created_at_nan(weibull_df: pandas.DataFrame) -> None:
    df = weibull_df
    df.loc[df.index[0], "created"] = None
    unit, groups, (G, B, T) = convoys.utils.get_arrays(df)
    assert numpy.issubdtype(G.dtype, numpy.integer)
    assert numpy.issubdtype(B.dtype, numpy.bool_)
    assert numpy.issubdtype(T.dtype, numpy.number)


@pytest.mark.parametrize(
    "model,extra_model",
    [
        ("kaplan-meier", None),
        ("weibull", None),
        ("exponential", None),
        ("kaplan-meier", "weibull"),
    ],
)
@flaky.flaky
def test_plot_cohorts(
    model: Literal[
        "kaplan-meier", "exponential", "weibull", "gamma", "generalized-gamma"
    ],
    extra_model: Literal[
        "kaplan-meier", "exponential", "weibull", "gamma", "generalized-gamma"
    ]
    | None,
    weibull_df: pandas.DataFrame,
) -> None:
    df = weibull_df
    unit, groups, (G, B, T) = convoys.utils.get_arrays(df)
    matplotlib.pyplot.clf()
    convoys.plotting.plot_cohorts(G, B, T, model=model, ci=0.95, groups=groups)
    matplotlib.pyplot.legend()
    if extra_model:
        convoys.plotting.plot_cohorts(
            G, B, T, model=extra_model, plot_kwargs=dict(linestyle="--", alpha=0.1)
        )
    here = Path(__file__)
    snapshots_dir = here.parent / "snapshots"
    snapshots_dir.mkdir(exist_ok=True)
    matplotlib.pyplot.savefig(
        snapshots_dir / f"{model}-{extra_model}.png"
        if extra_model is not None
        else snapshots_dir / f"{model}.png"
    )


def test_plot_cohorts_bad_model_raises(weibull_df: pandas.DataFrame) -> None:
    df = weibull_df
    unit, groups, (G, B, T) = convoys.utils.get_arrays(df)

    with pytest.raises(ValueError):
        convoys.plotting.plot_cohorts(G, B, T, model="bad", groups=groups)  # type: ignore[arg-type]


def test_plot_cohorts_bad_groups_raises(weibull_df: pandas.DataFrame) -> None:
    df = weibull_df
    unit, groups, (G, B, T) = convoys.utils.get_arrays(df)

    with pytest.raises(ValueError):
        convoys.plotting.plot_cohorts(
            G, B, T, model="kaplan-meier", groups=groups, specific_groups=["Nonsense"]
        )


def test_plot_cohorts_subplots(weibull_df: pandas.DataFrame) -> None:
    df = weibull_df
    unit, groups, (G, B, T) = convoys.utils.get_arrays(df)
    matplotlib.pyplot.clf()
    fix, axes = matplotlib.pyplot.subplots(nrows=2, ncols=2)
    for ax in axes.flatten():
        convoys.plotting.plot_cohorts(G, B, T, groups=groups, ax=ax)
        ax.legend()
    here = Path(__file__)
    snapshots_dir = here.parent / "snapshots"
    snapshots_dir.mkdir(exist_ok=True)
    matplotlib.pyplot.savefig(snapshots_dir / "subplots.png")


@pytest.fixture
def add_examples_to_path() -> None:
    import sys

    root = Path(__file__).parent.parent
    sys.path.append(str(root))


def test_marriage_example(add_examples_to_path: None) -> None:
    from examples.marriage import run

    run()


def test_dob_violations_example(add_examples_to_path: None) -> None:
    from examples.dob_violations import run

    run()



from __future__ import annotations

import datetime
from typing import TYPE_CHECKING, Any, Callable, Hashable, Literal, Sequence

import numpy
import pandas

if TYPE_CHECKING:
    Unit = Literal["years", "days", "hours", "minutes", "seconds"]

__all__ = ["get_arrays"]


def get_timescale(
    t: datetime.timedelta | pandas.Timedelta, unit: "Unit" | None
) -> tuple["Unit" | None, Callable[[Any], float]]:
    """Take a datetime or a numerical type, return two things:

    1. A unit
    2. A function that converts it to numerical form
    """

    def get_timedelta_converter(
        t_factor: int | float,
    ) -> Callable[[datetime.timedelta | pandas.Timedelta], float]:
        return lambda td: td.total_seconds() * t_factor

    if not isinstance(t, datetime.timedelta) or not isinstance(t, pandas.Timedelta):
        # Assume numeric type
        return None, lambda x: float(x)
    for u, f in [
        ("years", 365.25 * 24 * 60 * 60),
        ("days", 24 * 60 * 60),
        ("hours", 60 * 60),
        ("minutes", 60),
        ("seconds", 1),
    ]:
        if u == unit or (unit is None and t >= datetime.timedelta(seconds=f)):
            return u, get_timedelta_converter(1.0 / f)  # type: ignore[return-value]
    raise ValueError(f"Could not find unit for {t} and {unit}")


def get_groups(
    data: pandas.Series, group_min_size: int, max_groups: int
) -> list[Hashable]:
    """Picks the top groups out of a dataset

    1. Remove groups with too few data points
    2. Pick the top groups
    3. Sort groups lexicographically
    """
    group2count: dict[Hashable, int] = {}
    for group in data:
        group2count[group] = group2count.get(group, 0) + 1

    groups = [group for group, count in group2count.items() if count >= group_min_size]
    if max_groups >= 0:
        groups = sorted(
            groups,
            key=group2count.get,  # type: ignore[arg-type]
            reverse=True,
        )[:max_groups]
    return sorted(groups, key=lambda g: (g is None, g))  # Put Nones last


def _sub(a: datetime.datetime | Any, b: Any) -> Any:
    # Computes a - b for a bunch of different cases
    if isinstance(a, datetime.datetime) and a.tzinfo is not None:
        return a.astimezone(b.tzinfo) - b
    else:
        # Either naive timestamps or numerical type
        return a - b


def get_arrays(
    data: pandas.DataFrame,
    features: str | Sequence[str] | None = None,
    groups: str | None = None,
    created: str | None = None,
    converted: str | None = None,
    now: str | None = None,
    unit: "Unit" | None = None,
    group_min_size: int = 0,
    max_groups: int = -1,
) -> tuple[
    "Unit" | None,
    list[Hashable] | None,
    tuple[numpy.ndarray, numpy.ndarray, numpy.ndarray],
]:
    """Converts a dataframe to a list of numpy arrays.

    Generates either feature data, or group data.

    :param data: Pandas dataframe
    :param features: string or Sequence[str] (optional), refers to a column
        in the dataframe containing features, each being a 1d-vector or list
        of features. If not provided, then it it will look for a column in the
        dataframe named "features". This argument can also be a list of columns.
    :param groups: string (optional), refers to a column in the dataframe
        containing the groups for each row. If not provided, then it will
        look for a column in the dataframe named "groups".
    :param created: string (optional), refers to a column in the dataframe
        containing timestamps of when each item was "created". If not
        provided, then it will look for a column in the dataframe named
        "created".
    :param converted: string, refers to a column in the dataframe
        containing timestamps of when each item converted. If there is no
        column containing creation values, then the converted values should
        be timedeltas denoting time until conversion. If this argument is
        not provided, then it will look for a column in the dataframe named
        "created".
    :param now: string (optional), refers to a column in the dataframe
        containing the point in time up until which we have observed
        non-conversion. If there is no column containing creation value,
        then these values should be timedeltas. If this argument is not
        provided, the current timestamp will be used.
    :param unit: string (optional), time unit to use when converting to
        numerical values. Has to be one of "years", "days", "hours",
        "minutes", or "seconds". If not provided, then a choice will be
        made based on the largest time interval in the inputs.
    :param group_min_size: integer (optional), only include groups that
        has at least this many observations
    :param max_groups: integer (optional), only include the `n` largest
        groups
    :returns: tuple (unit, groups, arrays)

        `unit` is the unit chosen. Will be one of "years", "days", "hours",
        "minutes", or "seconds". If the `unit` parameter is passed, this
        will be the same.

        `groups` is a list of strings containing the groups. Will be `None`
         if `groups` is not set.

        `arrays` is a tuple of numpy arrays `(G, B, T)` or `(X, B, T)`
        containing the transformed input in numerical format. `G`, `B`, `T`
        will all be 1D numpy arrays. `X` will be a 2D numpy array.
    """
    # First, construct either the `X` or the `G` array
    if features is None and groups is None:
        if "group" in data.columns:
            groups = "group"
        elif "features" in data.columns:
            features = "features"
        else:
            raise Exception(
                "Neither of the `features` or `group` parameters"
                " was provided, and there was no `features` or"
                " `groups` dataframe column"
            )
    if groups is not None:
        groups_list = get_groups(data[groups], group_min_size, max_groups)
        group2j = dict((group, j) for j, group in enumerate(groups_list))
        # Remove rows for rare groups
        data = data[data[groups].isin(group2j.keys())]
        G = data[groups].apply(lambda g: group2j.get(g, -1)).to_numpy()
        retval = G
    else:
        groups_list = None
        if isinstance(features, tuple):
            features = list(features)  # Otherwise sad Panda
        # this creates an array of shape (n, k), whether features is a
        # single Series containing tuples of length k, or features is a
        # list of columns of length k.
        X = numpy.array([numpy.array(z) for z in data[features].values])
        retval = X

    # Next, construct the `B` and `T` arrays
    if converted is None:
        if "converted" in data.columns:
            converted = "converted"
        else:
            raise Exception(
                "The `converted` parameter was not provided"
                " and there was no `converted` dataframe column"
            )
    if now is None and "now" in data.columns:
        now = "now"
    if created is None and "created" in data.columns:
        created = "created"
    B = ~pandas.isnull(data[converted]).to_numpy()

    def _calculate_T(row: pandas.Series) -> Any:
        if not pandas.isnull(row[converted]):
            if created is not None:
                return _sub(row[converted], row[created])
            else:
                return row[converted]
        else:
            if created is not None:
                if now is not None:
                    return _sub(row[now], row[created])
                else:
                    return datetime.datetime.now(tz=row[created].tzinfo) - row[created]
            elif now is not None:
                return row[now]
            else:
                return datetime.datetime.now()

    T_deltas = data.apply(_calculate_T, axis=1)
    max_T_delta = T_deltas.max()
    unit, converter = get_timescale(max_T_delta, unit)
    T = T_deltas.apply(converter).to_numpy()

    return unit, groups_list, (retval, B, T)



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Callable, Generic, Hashable, TypeVar

import numpy

from convoys import regression, single

if TYPE_CHECKING:
    from numpy import ndarray
    from numpy.typing import ArrayLike

T_Group = TypeVar("T_Group", "ArrayLike", Hashable)

__all__ = ["KaplanMeier", "Exponential", "Weibull", "Gamma", "GeneralizedGamma"]


class MultiModel(ABC, Generic[T_Group]):
    _base_model_cls: type[regression.RegressionModel | single.SingleModel]

    @abstractmethod
    def fit(self, G: "ArrayLike", B: "ArrayLike", T: "ArrayLike") -> None:
        raise NotImplementedError("Need to implement fit")

    @abstractmethod
    def predict(self, group: T_Group, t: "ArrayLike") -> "ndarray":
        raise NotImplementedError("Need to implement predict")

    @abstractmethod
    def predict_ci(self, group: T_Group, t: "ArrayLike", ci: float) -> "ndarray":
        raise NotImplementedError("Need to implement predict_ci")


class RegressionToMulti(MultiModel["ArrayLike"]):
    _base_model_cls: type[regression.RegressionModel]

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self.base_model = self._base_model_cls(*args, **kwargs)

    def fit(self, G: "ArrayLike", B: "ArrayLike", T: "ArrayLike") -> None:
        """Fits the model

        :param G: numpy array of shape :math:`n`, containing integers representing group
            assignments.
        :param B: numpy array of shape :math:`n`, containing booleans representing
            whether or not the subject 'converted' at time delta :math:`t`.
        :param T: numpy array of shape :math:`n`, containing floats representing the
            time delta :math:`t` between creation and either conversion or censoring.
        """
        G = numpy.array(G, dtype=int)
        (n,) = G.shape
        self._n_groups = max(G) + 1
        X = numpy.zeros((n, self._n_groups), dtype=numpy.bool)
        for i, group in enumerate(G):
            X[i, group] = 1
        self.base_model.fit(X, B, T)

    def _get_x(self, group: "ArrayLike") -> "ndarray":
        x = numpy.zeros(self._n_groups)
        g = numpy.array(group)
        x[g] = 1
        return x

    def predict(self, group: "ArrayLike", t: "ArrayLike") -> "ndarray":
        return self.base_model.predict(self._get_x(group), t)

    def predict_ci(self, group: "ArrayLike", t: "ArrayLike", ci: float) -> "ndarray":
        return self.base_model.predict_ci(self._get_x(group), t, ci)

    def rvs(
        self, group: "ArrayLike", *args: Any, **kwargs: Any
    ) -> tuple["ndarray", "ndarray"]:
        return self.base_model.rvs(self._get_x(group), *args, **kwargs)


class SingleToMulti(MultiModel[Hashable]):
    _base_model_cls: type[single.SingleModel]

    def __init__(self, *args: Any, **kwargs: Any):
        self.base_model_init: Callable[[], single.SingleModel] = (
            lambda: self._base_model_cls(*args, **kwargs)
        )

    def fit(self, G: "ArrayLike", B: "ArrayLike", T: "ArrayLike") -> None:
        """Fits the model

        :param G: numpy array of shape :math:`n`, containing integers representing group
            assignments.
        :param B: numpy array of shape :math:`n`, containing booleans representing
            whether or not the subject 'converted' at time delta :math:`t`.
        :param T: numpy array of shape :math:`n`, containing floats representing the
            time delta :math:`t` between creation and either conversion or censoring.
        """
        G = numpy.array(G)
        B = numpy.array(B)
        T = numpy.array(T)
        group2bt: dict[Hashable, list[tuple]] = {}
        for g, b, t in zip(G, B, T, strict=False):
            group2bt.setdefault(g, []).append((b, t))
        self._group2model: dict[Hashable, single.SingleModel] = {}
        for g, BT in group2bt.items():
            self._group2model[g] = self.base_model_init()
            self._group2model[g].fit([b for b, t in BT], [t for b, t in BT])

    def predict(self, group: Hashable, t: "ArrayLike") -> "ndarray":
        return self._group2model[group].predict(t)

    def predict_ci(self, group: Hashable, t: "ArrayLike", ci: float) -> "ndarray":
        return self._group2model[group].predict_ci(t, ci)


class Exponential(RegressionToMulti):
    """Multi-group version of :class:`convoys.regression.Exponential`."""

    _base_model_cls = regression.Exponential


class Weibull(RegressionToMulti):
    """Multi-group version of :class:`convoys.regression.Weibull`."""

    _base_model_cls = regression.Weibull


class Gamma(RegressionToMulti):
    """Multi-group version of :class:`convoys.regression.Gamma`."""

    _base_model_cls = regression.Gamma


class GeneralizedGamma(RegressionToMulti):
    """Multi-group version of :class:`convoys.regression.GeneralizedGamma`."""

    _base_model_cls = regression.GeneralizedGamma


class KaplanMeier(SingleToMulti):
    """Multi-group version of :class:`convoys.single.KaplanMeier`."""

    _base_model_cls = single.KaplanMeier



import warnings
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

import numpy
import scipy.stats  # type: ignore[import-untyped]

if TYPE_CHECKING:
    from numpy.typing import ArrayLike

__all__ = ["KaplanMeier"]


class SingleModel(ABC):
    @abstractmethod
    def fit(self, B: "ArrayLike", T: "ArrayLike") -> None:
        raise NotImplementedError("Need to implement fit")

    @abstractmethod
    def predict(self, t: "ArrayLike") -> numpy.ndarray:
        raise NotImplementedError("Need to implement predict")

    @abstractmethod
    def predict_ci(self, t: "ArrayLike", ci: float) -> numpy.ndarray:
        raise NotImplementedError("Need to implement predict_ci")


class KaplanMeier(SingleModel):
    """Implementation of the Kaplan-Meier nonparametric method."""

    def fit(self, B: "ArrayLike", T: "ArrayLike") -> None:
        """Fits the model

        :param B: numpy array of shape :math:`n`, containing booleans representing
            whether or not the subject 'converted' at time delta :math:`t`.
        :param T: numpy array of shape :math:`n`, containing floats representing the
            time delta :math:`t` between creation and either conversion or censoring.
        """
        if not isinstance(B, numpy.ndarray):
            B = numpy.array(B)
        if not isinstance(T, numpy.ndarray):
            T = numpy.array(T)
        # See https://www.math.wustl.edu/~sawyer/handouts/greenwood.pdf
        BT = [
            (b, t) for b, t in zip(B, T, strict=False) if t >= 0 and 0 <= float(b) <= 1
        ]
        if len(BT) < len(B):
            n_removed = len(B) - len(BT)
            warnings.warn(
                "Warning! Removed %d/%d entries from inputs where "
                "T < 0 or B not 0/1" % (n_removed, len(B)),
                stacklevel=2,
            )
        B, T = ([z[i] for z in BT] for i in range(2))
        n = len(T)
        self._ts = [0.0]
        self._ss = [1.0]
        self._vs = [0.0]
        sum_var_terms = 0.0
        prod_s_terms = 1.0
        for t, b in sorted(zip(T, B, strict=False)):
            d = float(b)
            self._ts.append(t)
            prod_s_terms *= 1 - d / n
            self._ss.append(prod_s_terms)
            if d == n == 1:
                sum_var_terms = float("inf")
            else:
                sum_var_terms += d / (n * (n - d))
            if sum_var_terms > 0:
                self._vs.append(1 / numpy.log(prod_s_terms) ** 2 * sum_var_terms)
            else:
                self._vs.append(0)
            n -= 1

        # Just prevent overflow warning when computing the confidence interval
        eps = 1e-9
        self._ss_clipped = numpy.clip(self._ss, eps, 1.0 - eps)

    def predict(self, t: "ArrayLike") -> numpy.ndarray:
        """Returns the predicted values."""
        t = numpy.array(t)
        res = numpy.zeros(t.shape)
        for indexes, value in numpy.ndenumerate(t):
            j = numpy.searchsorted(self._ts, value, side="right") - 1
            if j >= len(self._ts) - 1:
                # Make the plotting stop at the last value of t
                res[indexes] = float("nan")
            else:
                res[indexes] = 1 - self._ss[j]
        return res

    def predict_ci(self, t: "ArrayLike", ci: float = 0.8) -> numpy.ndarray:
        """Returns the predicted values with a confidence interval."""
        t = numpy.array(t)
        res = numpy.zeros(t.shape + (3,))
        for indexes, value in numpy.ndenumerate(t):
            j = numpy.searchsorted(self._ts, value, side="right") - 1
            if j >= len(self._ts) - 1:
                # Make the plotting stop at the last value of t
                res[indexes] = [float("nan")] * 3
            else:
                z_lo, z_hi = scipy.stats.norm.ppf([(1 - ci) / 2, (1 + ci) / 2])
                res[indexes] = (
                    1 - self._ss[j],
                    1
                    - numpy.exp(
                        -numpy.exp(
                            numpy.log(-numpy.log(self._ss_clipped[j]))
                            + z_hi * self._vs[j] ** 0.5
                        )
                    ),
                    1
                    - numpy.exp(
                        -numpy.exp(
                            numpy.log(-numpy.log(self._ss_clipped[j]))
                            + z_lo * self._vs[j] ** 0.5
                        )
                    ),
                )
        return res



import datetime
from pathlib import Path

import pandas
from matplotlib import pyplot

import convoys.plotting
import convoys.utils

here = Path(__file__)


def run() -> None:
    print("loading data")
    df = pandas.read_pickle("examples/dob_violations.pickle")
    print(df["issue_date"])
    print(df["issue_date"].dtype)
    print(df["issue_date"] < datetime.date(2018, 1, 1))
    df = df[df["issue_date"] < datetime.date(2018, 1, 1)]

    print("converting to arrays")
    unit, groups, (G, B, T) = convoys.utils.get_arrays(
        df,
        groups="type",
        created="issue_date",
        converted="disposition_date",
        unit="years",
        group_min_size=100,
    )

    for model in ["kaplan-meier", "weibull"]:
        print("plotting", model)
        pyplot.figure(figsize=(9, 6))
        convoys.plotting.plot_cohorts(
            G,
            B,
            T,
            model=model,  # type: ignore[arg-type]
            ci=0.95,
            groups=groups,
            t_max=30,
        )
        pyplot.legend()
        assert unit is not None
        pyplot.xlabel(unit)
        fig_path = here.parent / f"dob-violations-{model}.png"
        pyplot.savefig(fig_path)

    pyplot.figure(figsize=(9, 6))
    df["bucket"] = df["issue_date"].apply(
        lambda d: "%d-%d" % (5 * (d.year // 5), 5 * (d.year // 5) + 4)
    )
    unit, groups, (G, B, T) = convoys.utils.get_arrays(
        df,
        groups="bucket",
        created="issue_date",
        converted="disposition_date",
        unit="years",
        group_min_size=500,
    )
    convoys.plotting.plot_cohorts(
        G, B, T, model="kaplan-meier", groups=groups, t_max=30, ci=0.95
    )
    convoys.plotting.plot_cohorts(
        G,
        B,
        T,
        model="weibull",
        groups=groups,
        t_max=30,
        plot_kwargs={"linestyle": "--"},
    )
    pyplot.legend()
    assert unit is not None
    pyplot.xlabel(unit)
    fig_path = here.parent / "dob-violations-combined.png"
    pyplot.savefig(fig_path)


if __name__ == "__main__":
    run()



from math import floor
from typing import Literal

import pandas
import pytest

import convoys.export
import convoys.utils


@pytest.mark.parametrize(
    "model",
    [
        "kaplan-meier",
        "weibull",
        "exponential",
    ],
)
def test_export_cohorts(
    model: Literal[
        "kaplan-meier", "exponential", "weibull", "gamma", "generalized-gamma"
    ],
    weibull_df: pandas.DataFrame,
) -> None:
    df = weibull_df
    unit, groups, (G, B, T) = convoys.utils.get_arrays(df, unit="days")
    assert groups is not None

    result_df = convoys.export.export_cohorts(
        G, B, T, model=model, ci=0.95, groups=groups
    )

    assert set(result_df["group"]) == set(groups)

    assert result_df["t"].dtype == int
    assert max(T) - 1 <= max(result_df["t"]) == floor(max(T))
    assert min(result_df["t"]) == 0

    assert 0 <= min(result_df["prediction_value"])
    assert max(result_df["prediction_value"]) <= 1


def test_export_cohorts_bad_model_raises(weibull_df: pandas.DataFrame) -> None:
    df = weibull_df
    unit, groups, (G, B, T) = convoys.utils.get_arrays(df)

    with pytest.raises(ValueError):
        convoys.export.export_cohorts(G, B, T, model="bad", groups=groups)  # type: ignore[arg-type]


def test_export_cohorts_bad_groups_raises(weibull_df: pandas.DataFrame) -> None:
    df = weibull_df
    unit, groups, (G, B, T) = convoys.utils.get_arrays(df)

    with pytest.raises(ValueError):
        convoys.export.export_cohorts(
            G, B, T, model="kaplan-meier", groups=groups, specific_groups=["Nonsense"]
        )



from __future__ import annotations

import warnings
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Callable, Literal

import emcee  # type: ignore[import-untyped]
import numpy
import progressbar
import scipy.optimize  # type: ignore[import-untyped]
from scipy.special import gammaincinv  # type: ignore[import-untyped]

with warnings.catch_warnings():
    # we have pinned numpy and SciPy, so we can ignore the deprecation warnings
    # issued by autograd 1.7
    warnings.filterwarnings("ignore", category=DeprecationWarning)
    import autograd  # type: ignore[import-untyped]
    from autograd.numpy import (  # type: ignore[import-untyped]
        dot,
        exp,
        isnan,
        log,
        sum,  # noqa: A004
    )
    from autograd.scipy.special import expit, gammaln  # type: ignore[import-untyped]
    from autograd_gamma import gammainc  # type: ignore[import-untyped]

if TYPE_CHECKING:
    from numpy.typing import ArrayLike


__all__ = ["Exponential", "Weibull", "Gamma", "GeneralizedGamma"]


def generalized_gamma_loss(
    x: tuple[float, ...],
    X: numpy.ndarray,
    B: numpy.ndarray,
    T: numpy.ndarray,
    W: numpy.ndarray,
    fix_k: int | None,
    fix_p: int | None,
    hierarchical: bool,
    flavor: Literal["logistic", "linear"],
    callback: Callable[[float], None] | None = None,
) -> float:
    k = exp(x[0]) if fix_k is None else fix_k
    p = exp(x[1]) if fix_p is None else fix_p
    log_sigma_alpha = x[2]
    log_sigma_beta = x[3]
    a = x[4]
    b = x[5]
    n_features = int((len(x) - 6) / 2)
    alpha = x[6 : 6 + n_features]
    beta = x[6 + n_features : 6 + 2 * n_features]
    lambd = exp(dot(X, alpha) + a)

    # PDF: p*lambda^(k*p) / gamma(k) * t^(k*p-1) * exp(-(x*lambda)^p)
    log_pdf = (
        log(p)
        + (k * p) * log(lambd)
        - gammaln(k)
        + (k * p - 1) * log(T)
        - (T * lambd) ** p
    )
    cdf = gammainc(k, (T * lambd) ** p)

    if flavor == "logistic":  # Log-likelihood with sigmoid
        c = expit(dot(X, beta) + b)
        LL_observed = log(c) + log_pdf
        LL_censored = log((1 - c) + c * (1 - cdf))
    elif flavor == "linear":  # L2 loss, linear
        c = dot(X, beta) + b
        LL_observed = -((1 - c) ** 2) + log_pdf
        LL_censored = -((c * cdf) ** 2)

    LL_data = sum(W * B * LL_observed + W * (1 - B) * LL_censored, 0)

    if hierarchical:
        # Hierarchical model with sigmas ~ invgamma(1, 1)
        LL_prior_a = (
            -4 * log_sigma_alpha
            - 1 / exp(log_sigma_alpha) ** 2
            - dot(alpha, alpha) / (2 * exp(log_sigma_alpha) ** 2)
            - n_features * log_sigma_alpha
        )
        LL_prior_b = (
            -4 * log_sigma_beta
            - 1 / exp(log_sigma_beta) ** 2
            - dot(beta, beta) / (2 * exp(log_sigma_beta) ** 2)
            - n_features * log_sigma_beta
        )
        LL: float = LL_prior_a + LL_prior_b + LL_data
    else:
        LL = LL_data

    if isnan(LL):
        return -numpy.inf
    if callback is not None:
        callback(LL)
    return LL


class RegressionModel(ABC):
    @abstractmethod
    def fit(
        self,
        X: "ArrayLike",
        B: "ArrayLike",
        T: "ArrayLike",
        W: "ArrayLike" | None = None,
    ) -> None:
        raise NotImplementedError("Need to implement fit")

    @abstractmethod
    def predict(self, group: "ArrayLike", t: "ArrayLike") -> numpy.ndarray:
        raise NotImplementedError("Need to implement predict")

    @abstractmethod
    def predict_ci(
        self, group: "ArrayLike", t: "ArrayLike", ci: float
    ) -> numpy.ndarray:
        raise NotImplementedError("Need to implement predict_ci")

    @abstractmethod
    def rvs(
        self, x: "ArrayLike", *args: Any, **kwargs: Any
    ) -> tuple[numpy.ndarray, numpy.ndarray]:
        raise NotImplementedError("Need to implement rvs")


class GeneralizedGamma(RegressionModel):
    """Generalization of Gamma, Weibull, and Exponential

    :param mcmc: boolean, defaults to False. Whether to use MCMC to
        sample from the posterior so that a confidence interval can be
        estimated later (see :meth:`predict`).
    :param fix_k: int or None. If an int, fixes :math:`k` to that value.
        For example, use `fix_k=1` for a Weibull distribution.
    :param fix_p: int or None. If an int, fixes :math:`p` to that value.
        Use `fix_p=1` for a Gamma distribution.
    :param hierarchical: boolean denoting whether we have a (Normal) prior
        on the alpha and beta parameters to regularize. The variance of
        the normal distribution is in itself assumed to be an inverse
        gamma distribution (1, 1).
    :param flavor: defaults to logistic. If set to 'linear', then an
        linear model is fit, where the beta params will be completely
        additive. This creates a much more interpretable model, with some
        minor loss of accuracy.

    This mostly follows the `Wikipedia article
    <https://en.wikipedia.org/wiki/Generalized_gamma_distribution>`_, although
    our notation is slightly different. Also see `this paper
    <https://grodri.github.io/survival/pop509slides1.pdf>`_ for an overview.

    **Shape of the probability function**

    The cumulative density function is:

    :math:`F(t) = P(k, (t\\lambda)^p)`

    where :math:`P(a, x) = \\gamma(a, x) / \\Gamma(a)` is the lower regularized
    incomplete gamma function.
    :math:`\\gamma(a, x)` is the incomplete gamma function and :math:`\\Gamma(a)`
    is the standard gamma function.

    The probability density function is:

    :math:`f(t) = p\\lambda^{kp} t^{kp-1} \\exp(-(t\\lambda)^p) / \\Gamma(k)`

    **Modeling conversion rate**

    Since our goal is to model the conversion rate, we assume the conversion
    rate converges to a final value

    :math:`c = \\sigma(\\mathbf{\\beta^Tx} + b)`

    where :math:`\\sigma(z) = 1/(1+e^{-z})` is the sigmoid function,
    :math:`\\mathbf{\\beta}` is an unknown vector we are solving for (with
    corresponding  intercept :math:`b`), and :math:`\\mathbf{x}` are the
    feature vector (inputs).

    We also assume that the rate parameter :math:`\\lambda` is determined by

    :math:`\\lambda = exp(\\mathbf{\\alpha^Tx} + a)`

    where :math:`\\mathrm{\\alpha}` is another unknown vector we are
    trying to solve for (with corresponding intercept :math:`a`).

    We also assume that the :math:`\\mathbf{\\alpha}, \\mathbf{\\beta}`
    vectors have a normal distribution

    :math:`\\alpha_i \\sim \\mathcal{N}(0, \\sigma_{\\alpha})`,
    :math:`\\beta_i \\sim \\mathcal{N}(0, \\sigma_{\\beta})`

    where hyperparameters :math:`\\sigma_{\\alpha}^2, \\sigma_{\\beta}^2`
    are drawn from an inverse gamma distribution

    :math:`\\sigma_{\\alpha}^2 \\sim \\text{inv-gamma}(1, 1)`,
    :math:`\\sigma_{\\beta}^2 \\sim \\text{inv-gamma}(1, 1)`

    **List of parameters**

    The full model fits vectors :math:`\\mathbf{\\alpha, \\beta}` and scalars
    :math:`a, b, k, p, \\sigma_{\\alpha}, \\sigma_{\\beta}`.

    **Likelihood and censorship**

    For entries that convert, the contribution to the likelihood is simply
    the probability density given by the probability distribution function
    :math:`f(t)` times the final conversion rate :math:`c`.

    For entries that *did not* convert, there are two options. Either the
    entry will never convert, which has probability :math:`1-c`. Or,
    it will convert at some later point that we have not observed yet,
    with probability given by the cumulative density function
    :math:`F(t)`.

    **Solving the optimization problem**

    To find the MAP (max a posteriori), `scipy.optimize.minimize
    <https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html#scipy.optimize.minimize>`_
    with the SLSQP method.

    If `mcmc == True`, then `emcee <http://dfm.io/emcee/current/>`_ is used
    to sample from the full posterior in order to generate uncertainty
    estimates for all parameters.
    """

    def __init__(
        self,
        mcmc: bool = False,
        fix_k: int | None = None,
        fix_p: int | None = None,
        hierarchical: bool = True,
        flavor: Literal["logistic", "linear"] = "logistic",
    ) -> None:
        self._mcmc = mcmc
        self._fix_k = fix_k
        self._fix_p = fix_p
        self._hierarchical = hierarchical
        self._flavor = flavor

    def fit(
        self,
        X: "ArrayLike",
        B: "ArrayLike",
        T: "ArrayLike",
        W: "ArrayLike" | None = None,
    ) -> None:
        """Fits the model.

        :param X: numpy 2-D array of shape :math:`k \\cdot n`, containing floats
            or ints representing the values of 1 or more features. Can be used
            with 1-hot encoding to represent group membership.
        :param B: numpy array of shape :math:`n`, containing booleans representing
            whether or not the subject 'converted' at time delta :math:`t`.
        :param T: numpy array of shape :math:`n`, containing floats representing the
            time delta :math:`t` between creation and either conversion or censoring.
        :param W: (optional) numpy vector of shape :math:`n`, containing floats
            representing weights (or counts) for the row's observation(s). If None,
            defaults to `numpy.ones(len(X))`.
        """

        if W is None:
            W = numpy.ones(len(X))  # type: ignore[arg-type]
        X, B, T, W = (
            Z if isinstance(Z, numpy.ndarray) else numpy.array(Z) for Z in (X, B, T, W)
        )
        keep_indexes = (T > 0) & (B >= 0) & (B <= 1) & (W >= 0)
        if sum(keep_indexes) < X.shape[0]:
            n_removed = X.shape[0] - sum(keep_indexes)
            warnings.warn(
                "Warning! Removed %d/%d entries from inputs where "
                "T <= 0 or B not 0/1 or W < 0" % (n_removed, len(X)),
                stacklevel=2,
            )
            X, B, T, W = (Z[keep_indexes] for Z in (X, B, T, W))
        n_features = X.shape[1]

        # scipy.optimize and emcee forces the the parameters to be a vector:
        # (log k, log p, log sigma_alpha, log sigma_beta,
        #  a, b, alpha_1...alpha_k, beta_1...beta_k)
        # Generalized Gamma is a bit sensitive to the starting point!
        x0 = numpy.zeros(6 + 2 * n_features)
        x0[0] = +1 if self._fix_k is None else log(self._fix_k)
        x0[1] = -1 if self._fix_p is None else log(self._fix_p)
        args = (X, B, T, W, self._fix_k, self._fix_p, self._hierarchical, self._flavor)

        # Set up progressbar and callback
        bar = progressbar.ProgressBar(
            widgets=[
                progressbar.Variable("loss", width=15, precision=9),  # type: ignore[no-untyped-call]
                " ",
                progressbar.BouncingBar(),  # type: ignore[no-untyped-call]
                " ",
                progressbar.Counter(width=6),  # type: ignore[no-untyped-call]
                " [",
                progressbar.Timer(),  # type: ignore[no-untyped-call]
                "]",
            ]
        )
        value_history = []

        def callback(LL: float) -> None:
            value_history.append(LL)
            bar.update(len(value_history), loss=LL)

        # Define objective and use automatic differentiation
        def f(x: tuple[float, ...]) -> float:
            return -generalized_gamma_loss(x, *args, callback=callback)

        jac = autograd.grad(lambda x: -generalized_gamma_loss(x, *args))

        # Find the maximum a posteriori of the distribution
        res = scipy.optimize.minimize(
            f, x0, jac=jac, method="SLSQP", options={"maxiter": 9999}
        )
        if not res.success:
            raise Exception("Optimization failed with message: %s" % res.message)
        result = {"map": res.x}

        # TODO: should not use fixed k/p as search parameters
        if self._fix_k:
            result["map"][0] = log(self._fix_k)
        if self._fix_p:
            result["map"][1] = log(self._fix_p)

        # Make sure we're in a local minimum
        gradient = jac(result["map"])
        gradient_norm = numpy.dot(gradient, gradient)
        if gradient_norm >= 1e-2 * len(X):
            warnings.warn(
                "Might not have found a local minimum! "
                "Norm of gradient is %f" % gradient_norm,
                stacklevel=2,
            )

        # Let's sample from the posterior to compute uncertainties
        if self._mcmc:
            (dim,) = res.x.shape
            n_walkers = 5 * dim
            sampler = emcee.EnsembleSampler(
                nwalkers=n_walkers,
                ndim=dim,
                log_prob_fn=generalized_gamma_loss,
                args=args,
            )
            mcmc_initial_noise = 1e-3
            p0 = [
                result["map"] + mcmc_initial_noise * numpy.random.randn(dim)
                for i in range(n_walkers)
            ]
            n_burnin = 100
            n_steps = int(numpy.ceil(2000.0 / n_walkers))
            n_iterations = n_burnin + n_steps

            bar = progressbar.ProgressBar(
                max_value=n_iterations,
                widgets=[
                    progressbar.Percentage(),  # type: ignore[no-untyped-call]
                    " ",
                    progressbar.Bar(),  # type: ignore[no-untyped-call]
                    " %d walkers [" % n_walkers,
                    progressbar.AdaptiveETA(),  # type: ignore[no-untyped-call]
                    "]",
                ],
            )
            for i, _ in enumerate(sampler.sample(p0, iterations=n_iterations)):
                bar.update(i + 1)
            result["samples"] = (
                sampler.get_chain()[n_burnin:, :, :].reshape(-1, dim, order="F").T
            )
            if self._fix_k:
                result["samples"][0, :] = log(self._fix_k)
            if self._fix_p:
                result["samples"][1, :] = log(self._fix_p)

        self.params = {
            k: {
                "k": exp(data[0]),
                "p": exp(data[1]),
                "a": data[4],
                "b": data[5],
                "alpha": data[6 : 6 + n_features].T,
                "beta": data[6 + n_features : 6 + 2 * n_features].T,
            }
            for k, data in result.items()
        }

    def _predict(
        self, params: dict[str, Any], x: "ArrayLike", t: "ArrayLike"
    ) -> numpy.ndarray:
        lambd = exp(dot(x, params["alpha"].T) + params["a"])
        if self._flavor == "logistic":
            c = expit(dot(x, params["beta"].T) + params["b"])
        elif self._flavor == "linear":
            c = dot(x, params["beta"].T) + params["b"]
        else:
            raise ValueError("flavor must be one of `logistic` or `linear`")
        M = c * gammainc(params["k"], (t * lambd) ** params["p"])

        return M  # type: ignore[no-any-return]

    def predict_posteriori(self, x: "ArrayLike", t: "ArrayLike") -> numpy.ndarray:
        """Returns the trace samples generated via the MCMC steps.

        Requires the model to be fit with `mcmc == True`."""
        x = numpy.array(x)
        t = numpy.array(t)
        assert self._mcmc
        params = self.params["samples"]
        t = numpy.expand_dims(t, -1)
        return self._predict(params, x, t)

    def predict_ci(
        self, x: "ArrayLike", t: "ArrayLike", ci: float = 0.8
    ) -> numpy.ndarray:
        """Works like :meth:`predict` but produces a confidence interval.

        Requires the model to be fit with `ci = True`. The return value
        will contain one more dimension than for :meth:`predict`, and
        the last dimension will have size 3, containing the mean, the
        lower bound of the confidence interval, and the upper bound of
        the confidence interval.
        """
        M = self.predict_posteriori(x, t)
        y = numpy.mean(M, axis=-1)
        y_lo = numpy.percentile(M, (1 - ci) * 50, axis=-1)
        y_hi = numpy.percentile(M, (1 + ci) * 50, axis=-1)
        return numpy.stack((y, y_lo, y_hi), axis=-1)

    def predict(self, x: "ArrayLike", t: "ArrayLike") -> numpy.ndarray:
        """Returns the value of the cumulative distribution function
        for a fitted model (using the maximum a posteriori estimate).

        :param x: feature vector (or matrix)
        :param t: time
        """
        params = self.params["map"]
        x = numpy.array(x)
        t = numpy.array(t)
        return self._predict(params, x, t)

    def rvs(
        self,
        x: "ArrayLike",
        n_curves: int = 1,
        n_samples: int = 1,
        T: numpy.ndarray | None = None,
    ) -> tuple[numpy.ndarray, numpy.ndarray]:
        """Samples values from this distribution

        T is optional and means we already observed non-conversion until T
        """
        assert self._mcmc  # Need to be fit with MCMC
        if T is None:
            T = numpy.zeros((n_curves, n_samples))
        else:
            assert T.shape == (n_curves, n_samples)
        B = numpy.zeros((n_curves, n_samples), dtype=numpy.bool)
        C = numpy.zeros((n_curves, n_samples))
        params = self.params["samples"]
        for i, j in enumerate(numpy.random.randint(len(params["k"]), size=n_curves)):
            k = params["k"][j]
            p = params["p"][j]
            lambd = exp(dot(x, params["alpha"][j]) + params["a"][j])
            c = expit(dot(x, params["beta"][j]) + params["b"][j])
            z = numpy.random.uniform(size=(n_samples,))
            cdf_now = c * gammainc(
                k, numpy.multiply.outer(T[i], lambd) ** p
            )  # why is this outer?
            adjusted_z = cdf_now + (1 - cdf_now) * z
            B[i] = adjusted_z < c
            y = adjusted_z / c
            w = gammaincinv(k, y)
            # x = (t * lambd)**p
            C[i] = w ** (1.0 / p) / lambd
            C[i][~B[i]] = 0

        return B, C


class Exponential(GeneralizedGamma):
    """Specialization of :class:`.GeneralizedGamma` where :math:`k=1, p=1`.

    The cumulative density function is:

    :math:`F(t) = 1 - \\exp(-t\\lambda)`

    The probability density function is:

    :math:`f(t) = \\lambda\\exp(-t\\lambda)`

    The exponential distribution is the most simple distribution.
    From a conversion perspective, you can interpret it as having
    two competing final states where the probability of transitioning
    from the initial state to converted or dead is constant.

    See documentation for :class:`GeneralizedGamma`."""

    def __init__(
        self,
        mcmc: bool = False,
        hierarchical: bool = True,
        flavor: Literal["logistic", "linear"] = "logistic",
    ) -> None:
        super().__init__(
            mcmc=mcmc, hierarchical=hierarchical, flavor=flavor, fix_p=1, fix_k=1
        )


class Weibull(GeneralizedGamma):
    """Specialization of :class:`.GeneralizedGamma` where :math:`k=1`.

    The cumulative density function is:

    :math:`F(t) = 1 - \\exp(-(t\\lambda)^p)`

    The probability density function is:

    :math:`f(t) = p\\lambda(t\\lambda)^{p-1}\\exp(-(t\\lambda)^p)`

    See documentation for :class:`GeneralizedGamma`."""

    def __init__(
        self,
        mcmc: bool = False,
        hierarchical: bool = True,
        flavor: Literal["logistic", "linear"] = "logistic",
    ) -> None:
        super().__init__(
            mcmc=mcmc, hierarchical=hierarchical, flavor=flavor, fix_k=1, fix_p=None
        )


class Gamma(GeneralizedGamma):
    """Specialization of :class:`.GeneralizedGamma` where :math:`p=1`.

    The cumulative density function is:

    :math:`F(t) = P(k, t\\lambda)`

    where :math:`P(a, x) = \\gamma(a, x) / \\Gamma(a)` is the lower regularized
    incomplete gamma function.

    The probability density function is:

    :math:`f(t) = \\lambda^k t^{k-1} \\exp(-x\\lambda) / \\Gamma(k)`

    See documentation for :class:`GeneralizedGamma`."""

    def __init__(
        self,
        mcmc: bool = False,
        hierarchical: bool = True,
        flavor: Literal["logistic", "linear"] = "logistic",
    ) -> None:
        super().__init__(
            mcmc=mcmc, hierarchical=hierarchical, flavor=flavor, fix_p=1, fix_k=None
        )
